/*
 * Copyright (c) 2017, NXP Semiconductor, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * o Redistributions of source code must retain the above copyright notice, this list
 *   of conditions and the following disclaimer.
 *
 * o Redistributions in binary form must reproduce the above copyright notice, this
 *   list of conditions and the following disclaimer in the documentation and/or
 *   other materials provided with the distribution.
 *
 * o Neither the name of NXP Semiconductor, Inc. nor the names of its
 *   contributors may be used to endorse or promote products derived from this
 *   software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * @file    Hexi_Project.c
 * @brief   Application entry point.
 */
#include <stdio.h>
#include "board.h"
#include "peripherals.h"
#include "pin_mux.h"
#include "clock_config.h"
#include "MK64F12.h"

#include "fsl_notifier.h"

#include "fsl_pmc.h"
#include "fsl_uart.h"
#include "fsl_common.h"
#include "fsl_smc.h"
#include "fsl_llwu.h"
#include "fsl_rcm.h"
#include "fsl_lptmr.h"
#include "fsl_port.h"

//#define MY_DEBUG

#ifdef MY_DEBUG
#include "fsl_debug_console.h"
#endif
/*******************************************************************************
 * Definitions
 ******************************************************************************/
#ifdef MY_DEBUG
#define APP_DEBUG_UART_BAUDRATE 		9600                 /* Debug console baud rate.           */
#define APP_DEBUG_UART_CLKSRC_NAME 		kCLOCK_CoreSysClk /* System clock.       */
#endif

#define LLWU_LPTMR_IDX 			0U       /* LLWU_M0IF */
#define LLWU_WAKEUP_PIN_IDX 	3U /* LLWU_P10  */
#define LLWU_WAKEUP_PIN_TYPE	kLLWU_ExternalPinFallingEdge

#define APP_WAKEUP_BUTTON_GPIO 			BOARD_SW2_GPIO
#define APP_WAKEUP_BUTTON_PORT 			BOARD_SW2_PORT
#define APP_WAKEUP_BUTTON_GPIO_PIN 		BOARD_SW2_GPIO_PIN
#define APP_WAKEUP_BUTTON_IRQ 			BOARD_SW2_IRQ
#define APP_WAKEUP_BUTTON_IRQ_HANDLER 	BOARD_SW2_IRQ_HANDLER
#define APP_WAKEUP_BUTTON_NAME 			BOARD_SW2_NAME
#define APP_WAKEUP_BUTTON_IRQ_TYPE 		kPORT_InterruptFallingEdge

#ifdef MY_DEBUG
/* Debug console RX pin: PORTB16 MUX: 3 */
#define DEBUG_CONSOLE_RX_PORT 			PORTB
#define DEBUG_CONSOLE_RX_GPIO 			GPIOB
#define DEBUG_CONSOLE_RX_PIN 			16
#define DEBUG_CONSOLE_RX_PINMUX 		kPORT_MuxAlt3
/* Debug console TX pin: PORTB17 MUX: 3 */
#define DEBUG_CONSOLE_TX_PORT 			PORTB
#define DEBUG_CONSOLE_TX_GPIO 			GPIOB
#define DEBUG_CONSOLE_TX_PIN 			17
#define DEBUG_CONSOLE_TX_PINMUX 		kPORT_MuxAlt3
#endif

#define CORE_CLK_FREQ 					CLOCK_GetFreq(kCLOCK_CoreSysClk)

/* Power mode definition used in application. */
typedef enum _app_power_mode {
	kAPP_PowerModeMin = 'A' - 1, kAPP_PowerModeRun, /*!< Run mode. All Kinetis chips. */
	kAPP_PowerModeWait, /*!< Wait mode. All Kinetis chips. */
	kAPP_PowerModeStop, /*!< Stop mode. All Kinetis chips. */
	kAPP_PowerModeVlpr, /*!< Very low power run mode. All Kinetis chips. */
	kAPP_PowerModeVlpw, /*!< Very low power wait mode. All Kinetis chips. */
	kAPP_PowerModeVlps, /*!< Very low power stop mode. All Kinetis chips. */
	kAPP_PowerModeLls, /*!< Low leakage stop mode. All Kinetis chips. */
	kAPP_PowerModeVlls0, /*!< Very low leakage stop 0 mode. Chip-specific. */
	kAPP_PowerModeVlls1, /*!< Very low leakage stop 1 mode. All Kinetis chips. */
	kAPP_PowerModeVlls2, /*!< Very low leakage stop 2 mode. All Kinetis chips. */
	kAPP_PowerModeVlls3, /*!< Very low leakage stop 3 mode. All Kinetis chips. */
	kAPP_PowerModeMax
} app_power_mode_t;

typedef struct power_user_config {
	app_power_mode_t mode; /*!< Power mode. */

	bool enableLowPowerWakeUpOnInterrupt; /*!< true - system exits to Run mode on interrupt from Very low power run,
	 Very low power wait or Very low power stop mode,
	 false - Wake-up on interrupt disabled for these modes. */
	bool enablePorDetectInVlls0; /*!< true - Power on reset detection circuit is enabled in Very low leakage stop 0
	 mode, false - Power on reset detection circuit is disabled. */

} power_user_config_t;

/* callback data type which is used for power manager user callback */
typedef struct {
	uint32_t beforeNotificationCounter; /*!< Callback before notification counter */
	uint32_t afterNotificationCounter; /*!< Callback after notification counter */
	smc_power_state_t originPowerState; /*!< Origin power state before switch */
} user_callback_data_t;

typedef enum _app_wakeup_source {
	kAPP_WakeupSourceLptmr, /*!< Wakeup by LPTMR.        */
	kAPP_WakeupSourcePin /*!< Wakeup by external pin. */
} app_wakeup_source_t;

/* TODO: insert other include files here. */

/* TODO: insert other definitions and declarations here. */

/*
 * Set the clock configuration for RUN mode from VLPR mode.
 */
extern void APP_SetClockRunFromVlpr(void);

/*
 * Set the clock configuration for VLPR mode.
 */
extern void APP_SetClockVlpr(void);

/*
 * Power mode switch callback.
 */
status_t callback0(notifier_notification_block_t *notify, void *dataPtr);

/*******************************************************************************
 * Variables
 ******************************************************************************/
static uint8_t s_wakeupTimeout; /* Wakeup timeout. (Unit: Second) */
static app_wakeup_source_t s_wakeupSource; /* Wakeup source.                 */

void APP_SetClockVlpr(void) {
	const sim_clock_config_t simConfig = { .pllFllSel = 3U, /* PLLFLLSEL select IRC48MCLK. */
	.er32kSrc = 2U, /* ERCLK32K selection, use RTC. */
	.clkdiv1 = 0x00040000U, /* SIM_CLKDIV1. */
	};

	CLOCK_SetSimSafeDivs();
	CLOCK_SetInternalRefClkConfig(kMCG_IrclkEnable, kMCG_IrcFast, 0U);

	/* MCG works in PEE mode now, will switch to BLPI mode. */

	CLOCK_ExternalModeToFbeModeQuick(); /* Enter FBE. */
	CLOCK_SetFbiMode(kMCG_Dmx32Default, kMCG_DrsLow, NULL); /* Enter FBI. */
	CLOCK_SetLowPowerEnable(true); /* Enter BLPI. */

	CLOCK_SetSimConfig(&simConfig);
}

void APP_SetClockRunFromVlpr(void) {
	const sim_clock_config_t simConfig = { .pllFllSel = 1U, /* PLLFLLSEL select PLL. */
	.er32kSrc = 2U, /* ERCLK32K selection, use RTC. */
	.clkdiv1 = 0x01140000U, /* SIM_CLKDIV1. */
	};

	const mcg_pll_config_t pll0Config = { .enableMode = 0U, .prdiv = 0x13U,
			.vdiv = 0x18U, };

	CLOCK_SetSimSafeDivs();

	/* Currently in BLPI mode, will switch to PEE mode. */
	/* Enter FBI. */
	CLOCK_SetLowPowerEnable(false);
	/* Enter FBE. */
	CLOCK_SetFbeMode(7U, kMCG_Dmx32Default, kMCG_DrsLow, NULL);
	/* Enter PBE. */
	CLOCK_SetPbeMode(kMCG_PllClkSelPll0, &pll0Config);
	/* Enter PEE. */
	CLOCK_SetPeeMode();

	CLOCK_SetSimConfig(&simConfig);
}

#ifdef MY_DEBUG
static void APP_InitDebugConsole(void) {
	uint32_t uartClkSrcFreq;
	uartClkSrcFreq = CLOCK_GetFreq(APP_DEBUG_UART_CLKSRC_NAME);
	DbgConsole_Init(BOARD_DEBUG_UART_BASEADDR, APP_DEBUG_UART_BAUDRATE,
	BOARD_DEBUG_UART_TYPE, uartClkSrcFreq);
}

#endif

status_t callback0(notifier_notification_block_t *notify, void *dataPtr) {
	user_callback_data_t *userData = (user_callback_data_t *) dataPtr;
	status_t ret = kStatus_Fail;
	app_power_mode_t targetMode =
			((power_user_config_t *) notify->targetConfig)->mode;
	smc_power_state_t originPowerState = userData->originPowerState;

	switch (notify->notifyType) {
	case kNOTIFIER_NotifyBefore:
		userData->beforeNotificationCounter++;
#ifdef MY_DEBUG
		/* Wait for debug console output finished. */
		while (!(kUART_TransmissionCompleteFlag
				& UART_GetStatusFlags((UART_Type *) BOARD_DEBUG_UART_BASEADDR))) {
		}
		DbgConsole_Deinit();
#endif
		if ((kAPP_PowerModeRun != targetMode)
				&& (kAPP_PowerModeVlpr != targetMode)) {
#ifdef MY_DEBUG
			/*
			 * Set pin for current leakage.
			 * Debug console RX pin: Set to pinmux to disable.
			 * Debug console TX pin: Don't need to change.
			 */
			PORT_SetPinMux(DEBUG_CONSOLE_RX_PORT, DEBUG_CONSOLE_RX_PIN, kPORT_PinDisabledOrAnalog);
#endif
		}

		ret = kStatus_Success;
		break;
	case kNOTIFIER_NotifyRecover:
		break;
	case kNOTIFIER_CallbackAfter:
		userData->afterNotificationCounter++;
		if ((kAPP_PowerModeRun != targetMode) && (kAPP_PowerModeVlpr != targetMode)) {
#ifdef MY_DEBUG
			/*
			 * Debug console RX pin is set to disable for current leakage, nee to re-configure pinmux.
			 * Debug console TX pin: Don't need to change.
			 */
			PORT_SetPinMux(DEBUG_CONSOLE_RX_PORT, DEBUG_CONSOLE_RX_PIN, DEBUG_CONSOLE_RX_PINMUX);
#endif
		}

		/*
		 * For some other platforms, if enter LLS mode from VLPR mode, when wakeup, the
		 * power mode is VLPR. But for some platforms, if enter LLS mode from VLPR mode,
		 * when wakeup, the power mode is RUN. In this case, the clock setting is still
		 * VLPR mode setting, so change to RUN mode setting here.
		 */
		if ((kSMC_PowerStateVlpr == originPowerState) && (kSMC_PowerStateRun == SMC_GetPowerModeState(SMC))) {
			APP_SetClockRunFromVlpr();
		}

		/*
		 * If enter stop modes when MCG in PEE mode, then after wakeup, the MCG is in PBE mode,
		 * need to enter PEE mode manually.
		 */
		if ((kAPP_PowerModeRun != targetMode)
				&& (kAPP_PowerModeWait != targetMode)
				&& (kAPP_PowerModeVlpw != targetMode)
				&& (kAPP_PowerModeVlpr != targetMode)) {
			if (kSMC_PowerStateRun == originPowerState) {
				/* Wait for PLL lock. */
				while (!(kMCG_Pll0LockFlag & CLOCK_GetStatusFlags())) {
				}
				CLOCK_SetPeeMode();
			}
		}
#ifdef MY_DEBUG
		APP_InitDebugConsole();
#endif
		ret = kStatus_Success;
		break;
	default:
		break;
	}
	return ret;
}

/*!
 * @brief LLWU interrupt handler.
 */
void LLWU_IRQHandler(void) {
	/* If wakeup by LPTMR. */
	if (LLWU_GetInternalWakeupModuleFlag(LLWU, LLWU_LPTMR_IDX)) {
		LPTMR_DisableInterrupts(LPTMR0, kLPTMR_TimerInterruptEnable);
		LPTMR_ClearStatusFlags(LPTMR0, kLPTMR_TimerCompareFlag);
		LPTMR_StopTimer(LPTMR0);
	}
	/* If wakeup by external pin. */
	if (LLWU_GetExternalWakeupPinFlag(LLWU, LLWU_WAKEUP_PIN_IDX)) {
		PORT_SetPinInterruptConfig(APP_WAKEUP_BUTTON_PORT,
		APP_WAKEUP_BUTTON_GPIO_PIN, kPORT_InterruptOrDMADisabled);
		PORT_ClearPinsInterruptFlags(APP_WAKEUP_BUTTON_PORT,
				(1U << APP_WAKEUP_BUTTON_GPIO_PIN));
		LLWU_ClearExternalWakeupPinFlag(LLWU, LLWU_WAKEUP_PIN_IDX);
	}
	/* Add for ARM errata 838869, affects Cortex-M4, Cortex-M4F Store immediate overlapping
	 exception return operation might vector to incorrect interrupt */
	__DSB();
}

/*!
 * @brief LPTMR0 interrupt handler.
 */
void LPTMR0_IRQHandler(void) {
	if (kLPTMR_TimerInterruptEnable & LPTMR_GetEnabledInterrupts(LPTMR0)) {
		LPTMR_DisableInterrupts(LPTMR0, kLPTMR_TimerInterruptEnable);
		LPTMR_ClearStatusFlags(LPTMR0, kLPTMR_TimerCompareFlag);
		LPTMR_StopTimer(LPTMR0);
	}
	/* Add for ARM errata 838869, affects Cortex-M4, Cortex-M4F Store immediate overlapping
	 exception return operation might vector to incorrect interrupt */
	__DSB();
}

/*!
 * @brief button interrupt handler.
 */
void APP_WAKEUP_BUTTON_IRQ_HANDLER(void) {
	if ((1U << APP_WAKEUP_BUTTON_GPIO_PIN)
			& PORT_GetPinsInterruptFlags(APP_WAKEUP_BUTTON_PORT)) {
		/* Disable interrupt. */
		PORT_SetPinInterruptConfig(APP_WAKEUP_BUTTON_PORT,
		APP_WAKEUP_BUTTON_GPIO_PIN, kPORT_InterruptOrDMADisabled);
		PORT_ClearPinsInterruptFlags(APP_WAKEUP_BUTTON_PORT,
				(1U << APP_WAKEUP_BUTTON_GPIO_PIN));
	}
	/* Add for ARM errata 838869, affects Cortex-M4, Cortex-M4F Store immediate overlapping
	 exception return operation might vector to incorrect interrupt */
	__DSB();
}

/*!
 * @brief Get input from user to obtain wakeup timeout
 */
static uint8_t APP_GetWakeupTimeout(void) {
	uint8_t timeout;

	while (1) {
#ifdef MY_DEBUG
		PRINTF("Select the wake up timeout in seconds.\r\n");
		PRINTF("The allowed range is 1s ~ 9s.\r\n");
		PRINTF("Eg. enter 5 to wake up in 5 seconds.\r\n");
		PRINTF("\r\nWaiting for input timeout value...\r\n\r\n");

		timeout = GETCHAR();

		PRINTF("%c\r\n", timeout);
		if ((timeout > '0') && (timeout <= '9')) {
			return timeout - '0';
		}
		PRINTF("Wrong value!\r\n");
#endif
	}

}

/*!
 * @brief Get wakeup source by user input.
 */
static app_wakeup_source_t APP_GetWakeupSource(void) {
	uint8_t ch;
	return kAPP_WakeupSourcePin;
//	while (1) {
//		PRINTF("Select the wake up source:\r\n");
//		PRINTF("Press T for LPTMR - Low Power Timer\r\n");
//		PRINTF("Press S for switch/button %s. \r\n", APP_WAKEUP_BUTTON_NAME);
//
//		PRINTF("\r\nWaiting for key press..\r\n\r\n");
//
//		ch = GETCHAR();
//
//		if ((ch >= 'a') && (ch <= 'z')) {
//			ch -= 'a' - 'A';
//		}
//
//		if (ch == 'T') {
//			return kAPP_WakeupSourceLptmr;
//		} else if (ch == 'S') {
//			return kAPP_WakeupSourcePin;
//		} else {
//			PRINTF("Wrong value!\r\n");
//		}
//	}
}

/*! @brief Get wakeup timeout and wakeup source. */
void APP_GetWakeupConfig(app_power_mode_t targetMode) {
	/* Get wakeup source by user input. */
	if (targetMode == kAPP_PowerModeVlls0) {
		/* In VLLS0 mode, the LPO is disabled, LPTMR could not work. */
#ifdef MY_DEBUG
		PRINTF("Not support LPTMR wakeup because LPO is disabled in VLLS0 mode.\r\n");
#endif
		s_wakeupSource = kAPP_WakeupSourcePin;
	} else {
		/* Get wakeup source by user input. */
		s_wakeupSource = APP_GetWakeupSource();
	}

	if (kAPP_WakeupSourceLptmr == s_wakeupSource) {
		/* Wakeup source is LPTMR, user should input wakeup timeout value. */
		s_wakeupTimeout = APP_GetWakeupTimeout();
#ifdef MY_DEBUG
		PRINTF("Will wakeup in %d seconds.\r\n", s_wakeupTimeout);
#endif
	} else {
#ifdef MY_DEBUG
		PRINTF("Press %s to wake up.\r\n", APP_WAKEUP_BUTTON_NAME);
#endif
	}
}

/*! @brief Set wakeup timeout and wakeup source. */
void APP_SetWakeupConfig(app_power_mode_t targetMode) {
	/* Set LPTMR timeout value. */
	if (kAPP_WakeupSourceLptmr == s_wakeupSource) {
		LPTMR_SetTimerPeriod(LPTMR0, (LPO_CLK_FREQ * s_wakeupTimeout) - 1U);
		LPTMR_StartTimer(LPTMR0);
	}

	/* Set the wakeup module. */
	if (kAPP_WakeupSourceLptmr == s_wakeupSource) {
		LPTMR_EnableInterrupts(LPTMR0, kLPTMR_TimerInterruptEnable);
	} else {
		PORT_SetPinInterruptConfig(APP_WAKEUP_BUTTON_PORT,
		APP_WAKEUP_BUTTON_GPIO_PIN, APP_WAKEUP_BUTTON_IRQ_TYPE);
	}

	/* If targetMode is VLLS/LLS, setup LLWU. */
	if ((kAPP_PowerModeWait != targetMode) && (kAPP_PowerModeVlpw != targetMode)
			&& (kAPP_PowerModeVlps != targetMode)
			&& (kAPP_PowerModeStop != targetMode)) {
		if (kAPP_WakeupSourceLptmr == s_wakeupSource) {
			LLWU_EnableInternalModuleInterruptWakup(LLWU, LLWU_LPTMR_IDX, true);
		} else {
			LLWU_SetExternalWakeupPinMode(LLWU, LLWU_WAKEUP_PIN_IDX,
			LLWU_WAKEUP_PIN_TYPE);
		}
		NVIC_EnableIRQ(LLWU_IRQn);
	}
}

/*! @brief Show current power mode. */
void APP_ShowPowerMode(smc_power_state_t currentPowerState) {
	switch (currentPowerState) {
	case kSMC_PowerStateRun:
#ifdef MY_DEBUG
		PRINTF("    Power mode: RUN\r\n");
#endif
		break;
	case kSMC_PowerStateVlpr:
#ifdef MY_DEBUG
		PRINTF("    Power mode: VLPR\r\n");
#endif
		break;
	default:
#ifdef MY_DEBUG
		PRINTF("    Power mode wrong\r\n");
#endif
		break;
	}
}

/*!
 * @brief check whether could switch to target power mode from current mode.
 * Return true if could switch, return false if could not switch.
 */
bool APP_CheckPowerMode(smc_power_state_t currentPowerState,
		app_power_mode_t targetPowerMode) {
	bool modeValid = true;

	/*
	 * Check wether the mode change is allowed.
	 *
	 * 1. If current mode is HSRUN mode, the target mode must be RUN mode.
	 * 2. If current mode is RUN mode, the target mode must not be VLPW mode.
	 * 3. If current mode is VLPR mode, the target mode must not be HSRUN/WAIT/STOP mode.
	 * 4. If already in the target mode, don't need to change.
	 */
	switch (currentPowerState) {
	case kSMC_PowerStateRun:
		if (kAPP_PowerModeVlpw == targetPowerMode) {
#ifdef MY_DEBUG
			PRINTF("Could not enter VLPW mode from RUN mode.\r\n");
#endif
			modeValid = false;
		}
		break;

	case kSMC_PowerStateVlpr:
		if ((kAPP_PowerModeWait == targetPowerMode)	|| (kAPP_PowerModeStop == targetPowerMode)) {
#ifdef MY_DEBUG
			PRINTF("Could not enter HSRUN/STOP/WAIT modes from VLPR mode.\r\n");
#endif
			modeValid = false;
		}
		break;
	default:
#ifdef MY_DEBUG
		PRINTF("Wrong power state.\r\n");
#endif
		modeValid = false;
		break;
	}

	if (!modeValid) {
		return false;
	}

	/* Don't need to change power mode if current mode is already the target mode. */
	if (((kAPP_PowerModeRun == targetPowerMode) && (kSMC_PowerStateRun == currentPowerState))
			|| ((kAPP_PowerModeVlpr == targetPowerMode) && (kSMC_PowerStateVlpr == currentPowerState))) {
#ifdef MY_DEBUG
		PRINTF("Already in the target power mode.\r\n");
#endif
		return false;
	}

	return true;
}

/*!
 * @brief Power mode switch.
 * This function is used to register the notifier_handle_t struct's member userFunction.
 */
status_t APP_PowerModeSwitch(notifier_user_config_t *targetConfig, void *userData) {
	smc_power_state_t currentPowerMode; /* Local variable with current power mode */
	app_power_mode_t targetPowerMode; /* Local variable with target power mode name*/
	power_user_config_t *targetPowerModeConfig; /* Local variable with target power mode configruation */

	smc_power_mode_vlls_config_t vlls_config; /* Local variable for vlls configuration */

	targetPowerModeConfig = (power_user_config_t *) targetConfig;
	currentPowerMode = SMC_GetPowerModeState(SMC);
	targetPowerMode = targetPowerModeConfig->mode;

	switch (targetPowerMode) {
	case kAPP_PowerModeVlpr:
		APP_SetClockVlpr();
		SMC_SetPowerModeVlpr(SMC, targetPowerModeConfig->enableLowPowerWakeUpOnInterrupt);
		while (kSMC_PowerStateVlpr != SMC_GetPowerModeState(SMC)) {
		}
		break;

	case kAPP_PowerModeRun:

		/* Power mode change. */
		SMC_SetPowerModeRun(SMC);
		while (kSMC_PowerStateRun != SMC_GetPowerModeState(SMC)) {
		}

		/* If enter RUN from VLPR, change clock after the power mode change. */
		if (kSMC_PowerStateVlpr == currentPowerMode) {
			APP_SetClockRunFromVlpr();
		}
		break;

		/* For wait modes. */
	case kAPP_PowerModeWait:
		SMC_PreEnterWaitModes();
		SMC_SetPowerModeWait(SMC);
		SMC_PostExitWaitModes();
		break;
	case kAPP_PowerModeVlpw:
		SMC_PreEnterWaitModes();
		SMC_SetPowerModeVlpw(SMC);
		SMC_PostExitWaitModes();
		break;

		/* For stop modes. */
	case kAPP_PowerModeStop:
		SMC_PreEnterStopModes();
		SMC_SetPowerModeStop(SMC, kSMC_PartialStop);
		SMC_PostExitStopModes();
		break;

	case kAPP_PowerModeVlps:
		SMC_PreEnterStopModes();
		SMC_SetPowerModeVlps(SMC);
		SMC_PostExitStopModes();
		break;

	case kAPP_PowerModeLls:
		SMC_PreEnterStopModes();
		SMC_SetPowerModeLls(SMC);
		SMC_PostExitStopModes();
		break;

	case kAPP_PowerModeVlls0:
	case kAPP_PowerModeVlls1:
	case kAPP_PowerModeVlls2:
	case kAPP_PowerModeVlls3:
		if (targetPowerMode == kAPP_PowerModeVlls3) {
			vlls_config.subMode = kSMC_StopSub3;
		} else if (targetPowerMode == kAPP_PowerModeVlls2) {
			vlls_config.subMode = kSMC_StopSub2;
		} else if (targetPowerMode == kAPP_PowerModeVlls0) {
			vlls_config.subMode = kSMC_StopSub0;
		} else {
			vlls_config.subMode = kSMC_StopSub1;
		}
		vlls_config.enablePorDetectInVlls0 = targetPowerModeConfig->enablePorDetectInVlls0;
		SMC_PreEnterStopModes();
		SMC_SetPowerModeVlls(SMC, &vlls_config);
		SMC_PostExitStopModes();
		break;
	default:


		//PRINTF("Wrong value");

		break;
	}
	return kStatus_Success;
}

/*
 * @brief   Application entry point.
 */
int main(void) {

	uint32_t freq = 0;
	uint8_t ch;
	uint8_t targetConfigIndex;
	notifier_handle_t powerModeHandle;
	smc_power_state_t currentPowerState;
	app_power_mode_t targetPowerMode;
	bool needSetWakeup; /* Flag of whether or not need to set wakeup. */
	lptmr_config_t lptmrConfig;

	/*Power mode configurations*/
	power_user_config_t vlprConfig = {
			kAPP_PowerModeVlpr,
			false,
			true,
	};

	power_user_config_t vlpwConfig = vlprConfig;
	power_user_config_t vlpsConfig = vlprConfig;
	power_user_config_t waitConfig = vlprConfig;
	power_user_config_t stopConfig = vlprConfig;
	power_user_config_t runConfig = vlprConfig;

	power_user_config_t llsConfig = vlprConfig;

	power_user_config_t vlls0Config = vlprConfig;
	power_user_config_t vlls1Config = vlprConfig;
	power_user_config_t vlls3Config = vlprConfig;
	power_user_config_t vlls2Config = vlprConfig;

	/* Initializes array of pointers to power mode configurations */
	notifier_user_config_t *powerConfigs[] = {
			&runConfig,
			&waitConfig,
			&stopConfig,
			&vlprConfig,
			&vlpwConfig,
			&vlpsConfig,
			&llsConfig,
			&vlls0Config,
			&vlls1Config,
			&vlls2Config,
			&vlls3Config,
	};

	/* User callback data0 */
	user_callback_data_t callbackData0;

	/* Initializes callback configuration structure */
	notifier_callback_config_t callbackCfg0 = {
			callback0,
			kNOTIFIER_CallbackBeforeAfter,
			(void *) &callbackData0 };

	/* Initializes array of callback configurations */
	notifier_callback_config_t callbacks[] = {
			callbackCfg0
	};

	memset(&callbackData0, 0, sizeof(user_callback_data_t));

	/* Initializes configuration structures */

	waitConfig.mode = kAPP_PowerModeWait;
	stopConfig.mode = kAPP_PowerModeStop;
	runConfig.mode = kAPP_PowerModeRun;

	vlpwConfig.mode = kAPP_PowerModeVlpw;
	vlpsConfig.mode = kAPP_PowerModeVlps;

	llsConfig.mode = kAPP_PowerModeLls;
	vlls0Config.mode = kAPP_PowerModeVlls0;
	vlls2Config.mode = kAPP_PowerModeVlls2;
	vlls1Config.mode = kAPP_PowerModeVlls1;
	vlls3Config.mode = kAPP_PowerModeVlls3;

	/* Create Notifier Handle */
	NOTIFIER_CreateHandle(
			&powerModeHandle,
			powerConfigs,
			ARRAY_SIZE(powerConfigs),
			callbacks,
			1U,
			APP_PowerModeSwitch,
			NULL);

	/* Power related. */
	SMC_SetPowerModeProtection(SMC, kSMC_AllowPowerModeAll);
	if (kRCM_SourceWakeup & RCM_GetPreviousResetSources(RCM)) /* Wakeup from VLLS. */
	{
		PMC_ClearPeriphIOIsolationFlag(PMC);
		NVIC_ClearPendingIRQ(LLWU_IRQn);
	}

	/* Init board hardware. */
	BOARD_InitBootPins();
	BOARD_InitBootClocks();
	//BOARD_InitBootPeripherals();
	/* Init FSL debug console. */
	//BOARD_InitDebugConsole();
#ifdef MY_DEBUG
	APP_InitDebugConsole();
#endif
	//printf("Hello World\n");

	/* Setup LPTMR. */
	/*
	 * lptmrConfig.timerMode = kLPTMR_TimerModeTimeCounter;
	 * lptmrConfig.pinSelect = kLPTMR_PinSelectInput_0;
	 * lptmrConfig.pinPolarity = kLPTMR_PinPolarityActiveHigh;
	 * lptmrConfig.enableFreeRunning = false;
	 * lptmrConfig.bypassPrescaler = true;
	 * lptmrConfig.prescalerClockSource = kLPTMR_PrescalerClock_1;
	 * lptmrConfig.value = kLPTMR_Prescale_Glitch_0;
	 */
	LPTMR_GetDefaultConfig(&lptmrConfig);
	/* Use LPO as clock source. */
	lptmrConfig.prescalerClockSource = kLPTMR_PrescalerClock_1;
	lptmrConfig.bypassPrescaler = true;

	LPTMR_Init(LPTMR0, &lptmrConfig);

	NVIC_EnableIRQ(LLWU_IRQn);
	NVIC_EnableIRQ(LPTMR0_IRQn);
	NVIC_EnableIRQ(APP_WAKEUP_BUTTON_IRQ);

	/* Wakeup from VLLS. */
	if (kRCM_SourceWakeup & RCM_GetPreviousResetSources(RCM)) {
#ifdef MY_DEBUG
		PRINTF("\r\nMCU wakeup from VLLS modes...\r\n");
#endif
	}

	/* Force the counter to be placed into memory. */
	//volatile static int i = 0;
	/* Enter an infinite loop, just incrementing a counter. */
	while (1) {
		//i++;

		currentPowerState = SMC_GetPowerModeState(SMC);
		freq = CLOCK_GetFreq(kCLOCK_CoreSysClk);
#ifdef MY_DEBUG
		PRINTF("\r\n####################  Power Manager Demo ####################\n\r\n");
		PRINTF("  Core Clock = %dHz \r\n", freq);
		APP_ShowPowerMode(currentPowerState);
#endif



//		PRINTF("\r\nSelect the desired operation \n\r\n");
//		PRINTF("Press  %c for enter: RUN      - Normal RUN mode\r\n",kAPP_PowerModeRun);
//		PRINTF("Press  %c for enter: WAIT     - Wait mode\r\n",kAPP_PowerModeWait);
//		PRINTF("Press  %c for enter: STOP     - Stop mode\r\n",kAPP_PowerModeStop);
//		PRINTF("Press  %c for enter: VLPR     - Very Low Power Run mode\r\n",kAPP_PowerModeVlpr);
//		PRINTF("Press  %c for enter: VLPW     - Very Low Power Wait mode\r\n",kAPP_PowerModeVlpw);
//		PRINTF("Press  %c for enter: VLPS     - Very Low Power Stop mode\r\n",kAPP_PowerModeVlps);
//		PRINTF("Press  %c for enter: LLS/LLS3 - Low Leakage Stop mode\r\n",kAPP_PowerModeLls);
//		PRINTF("Press  %c for enter: VLLS0    - Very Low Leakage Stop 0 mode\r\n",kAPP_PowerModeVlls0);
//		PRINTF("Press  %c for enter: VLLS1    - Very Low Leakage Stop 1 mode\r\n",kAPP_PowerModeVlls1);
//		PRINTF("Press  %c for enter: VLLS2    - Very Low Leakage Stop 2 mode\r\n",kAPP_PowerModeVlls2);
//		PRINTF("Press  %c for enter: VLLS3    - Very Low Leakage Stop 3 mode\r\n",kAPP_PowerModeVlls3);
//		PRINTF("\r\nWaiting for power mode select..\r\n\r\n");

		/* Wait for user response */
//		ch = GETCHAR();
//
//		if ((ch >= 'a') && (ch <= 'z')) {
//			ch -= 'a' - 'A';
//		}

		targetPowerMode = kAPP_PowerModeVlps; // (app_power_mode_t) ch; kAPP_PowerModeVlps

		if ((targetPowerMode > kAPP_PowerModeMin) && (targetPowerMode < kAPP_PowerModeMax)) {
			/* If could not set the target power mode, loop continue. */
			if (!APP_CheckPowerMode(currentPowerState, targetPowerMode)) {
				continue;
			}

			/* If target mode is RUN/VLPR/HSRUN, don't need to set wakeup source. */
			if ((kAPP_PowerModeRun == targetPowerMode) || (kAPP_PowerModeVlpr == targetPowerMode)) {
				needSetWakeup = false;
			} else {
				needSetWakeup = true;
			}

			if (needSetWakeup) {
				APP_GetWakeupConfig(targetPowerMode);
				//APP_PowerPreSwitchHook(curPowerState, targetPowerMode);
				APP_SetWakeupConfig(targetPowerMode);
			}

			callbackData0.originPowerState = currentPowerState;
			targetConfigIndex = targetPowerMode - kAPP_PowerModeMin - 1;
			NOTIFIER_SwitchConfig(&powerModeHandle, targetConfigIndex, kNOTIFIER_PolicyAgreement);
			//PRINTF("\r\nNext loop\r\n");
		}
	}
	//return 0;
}
